#  New menu funcions
#  Now menus are identified by UNTRANSLATED name (English is considered as reference untranslated string)
#  (positions of options in their menus start at 0)

namespace eval GiDMenu {

}

# Updates changes
proc GiDMenu::UpdateMenus {} {
    CreateTopMenus
}

# -----------------------------------------------------------------------------
#  GiDMenu::Delete
#
#  Delete a toplevel menu
#  menu_name_untranslated (untranslated english menu name, for example "Files#C#menu"
#  prepost: PRE POST or PREPOST
#  translationfunc can be _ (the default value, for GiD standard strings) or = (for module owned strings)
#  for example:
#    GiDMenu::Delete "Mesh#C#menu" PRE
#    it's also valid the old sintax, without #C#menu , but is more expensive to find, e.g:
#    GiDMenu::Delete "Mesh" PRE
#
#  Remember call GiDMenu::UpdateMenus to apply all changes.
# -----------------------------------------------------------------------------
proc GiDMenu::Delete { menu_name_untranslated prepost {translationfunc _} } {
    if {$prepost!="PRE" && $prepost!="POST" && $prepost!="PREPOST"} {
	return -code error [_ "Wrong arg. Must be PRE, POST or PREPOST."]
    }
    if { $prepost eq "PRE" } {
	set where [list "" "PRE"]
    } elseif { $prepost eq "POST" } {
	set where [list "P" "POST"]
    } else {
	set where [list "" "PRE" "P" "POST"]
    }
    foreach { c desc } $where {
	set i [GiDMenu::_FindIndex $menu_name_untranslated $desc $translationfunc]
	if { $i == "-1" } {
	    WarnWin [_ "ERROR: Menu %s doesn't exist in $desc." $menu_name_untranslated]
	    return
	}
	upvar \#0 MenuNames$c MenuNames
	upvar \#0 MenuEntries$c MenuEntries
	upvar \#0 MenuCommands$c MenuCommands
	upvar \#0 MenuAcceler$c MenuAcceler
	upvar \#0 MenuIcons$c MenuIcons
	
	set MenuNames [lreplace $MenuNames $i $i]
	array unset MenuEntries $i,*
	array unset MenuCommands $i,*
	array unset MenuAcceler $i,*
	array unset MenuIcons $i,*
	
	set l [llength $MenuNames]
	for {incr i} {$i <= $l} {incr i} {
	    foreach item [array names MenuEntries ${i}*] {
		set new_item [regsub ${i}* $item [expr $i-1]]
		set MenuEntries($new_item) $MenuEntries($item)
		unset MenuEntries($item)
	    }
	    foreach item [array names MenuCommands ${i}*] {
		set new_item [regsub ${i}* $item [expr $i-1]]
		set MenuCommands($new_item) $MenuCommands($item)
		unset MenuCommands($item)
	    }
	    foreach item [array names MenuAcceler ${i}*] {
		set new_item [regsub ${i}* $item [expr $i-1]]
		set MenuAcceler($new_item) $MenuAcceler($item)
		unset MenuAcceler($item)
	    }
	    foreach item [array names MenuIcons ${i}*] {
		set new_item [regsub ${i}* $item [expr $i-1]]
		set MenuIcons($new_item) $MenuIcons($item)
		unset MenuIcons($item)
	    }
	}
	array unset MenuEntries $l,*
	array unset MenuCommands $l,*
	array unset MenuAcceler $l,*
	array unset MenuIcons $l,*
	array unset MenuEntries $l
	array unset MenuCommands $l
	array unset MenuAcceler $l
	array unset MenuIcons $l
    }
}

# -----------------------------------------------------------------------------
#  GiDMenu::Create
#
#  Creates a new menu. New menus are inserted between the 'Calculate' and 'Help' menu
#  menu_name_untranslated (untranslated english menu name, for example "Files#C#menu"
#  prepost: PRE POST or PREPOST
#  translationfunc can be _ (the default value,for GiD standard strings) or = (for module strings)
#  pos: index where the new menu will be inserted, begin from 0, "end" can be used
#       if set to -1 (default) then is inserted before the 'Help' menu
#  Remember call GiDMenu::UpdateMenus to apply all changes.
# -----------------------------------------------------------------------------
proc GiDMenu::Create { menu_name_untranslated prepost {pos -1} {translationfunc _}} {
    if { $prepost != "PRE" && $prepost != "POST" && $prepost != "PREPOST" } {
	WarnWin [_ "Wrong arg. Must be PRE, POST or PREPOST."]
	return 1
    }
    if {$prepost eq "PRE"} {
	set where [list ""]
    } elseif {$prepost eq "POST"} {
	set where [list "P"]
    } else {
	set where [list "" "P"]
    }
    foreach c $where {
	upvar \#0 MenuNames$c MenuNames
	upvar \#0 MenuEntries$c MenuEntries
	upvar \#0 MenuCommands$c MenuCommands
	upvar \#0 MenuAcceler$c MenuAcceler
	upvar \#0 MenuIcons$c MenuIcons
	
	set length [llength $MenuNames]
	if { $pos == "end" } {
	    set pos $length
	}
	if { $pos == -1 } {
	    #find the position of the help menu
	    set pos [GiDMenu::_FindIndex "Help#C#menu" PRE _]
	    if { $pos == -1 } {
		set pos [expr [llength $MenuNames]-1]
	    }
	}
	if { $pos < 0 } {
	    set pos 0
	} elseif { $pos > $length } {
	    set pos $length
	}
	set MenuNames [linsert $MenuNames $pos [$translationfunc $menu_name_untranslated]]
	for { set i $length} { $i > $pos } { incr i -1 } {
	    set iprev [expr {$i-1}]
	    set MenuEntries($i) $MenuEntries($iprev)
	    set MenuCommands($i) $MenuCommands($iprev)
	    set MenuAcceler($i)  $MenuAcceler($iprev)
	    set MenuIcons($i) $MenuIcons($iprev)
	}
	set MenuEntries($pos)  [list ""]
	set MenuCommands($pos) [list ""]
	set MenuAcceler($pos)  [list ""]
	set MenuIcons($pos)  [list ""]
    }
    return 0
}


# -----------------------------------------------------------------------------
#  GiDMenu::InsertOption
#
#  Creates a new option for a given menu in a given  position.
#  menu_name_untranslated: the toplevel menu name untranslated
#  option_name_untranslated: the option name, can be a list for sublevels
#  e.g [list "Create#C#menu" "Arc#C#menu"]
#  it is also accepted the old name without comments: [list "Create" "Arc"]
#  position: the new menu option is inserted after the given position
#  (positions start at 0), "end" can be used
#  prepost: PRE POST or PREPOST
#  Arg 'command' is the command called when the menu option is selected.
#  Arg acceler is an optional key accelerator, like "Control-s"
#  Arg icon is the name of an small icon (16x16 pixels)
#  ins_repl: "replace" (default): the new menu option will replace an old if exists in the given position,
#            "insert" : will be inserted before the position
#            "insertafter" : will be inserted after the position
#  translationfunc can be _ (the default value,for GiD standard strings) or = (for module strings)
#
#  Remember call GiDMenu::UpdateMenus to apply all changes.
# -----------------------------------------------------------------------------

proc GiDMenu::InsertOption { menu_name_untranslated option_name_untranslated position prepost \
    command {acceler ""} {icon ""} {ins_repl "replace"} {translationfunc _} } {

    if {$prepost!="PRE" && $prepost!="POST" && $prepost!="PREPOST"} {
	WarnWin [_ "Wrong arg. Must be PRE, POST or PREPOST."]
	return 1
    }

    if {$prepost eq "PRE"} {
	set where [list "" "PRE"]
    } elseif {$prepost eq "POST"} {
	set where [list "P" "POST"]
    } else {
	set where [list "" "PRE" "P" "POST"]
    }
    foreach {c desc} $where {
	upvar \#0 MenuNames$c MenuNames
	upvar \#0 MenuEntries$c MenuEntries
	upvar \#0 MenuCommands$c MenuCommands
	upvar \#0 MenuAcceler$c MenuAcceler
	upvar \#0 MenuIcons$c MenuIcons
	
	set i [GiDMenu::_FindIndex $menu_name_untranslated $desc $translationfunc]
	if { $i == "-1" } {
	    WarnWin [_ "ERROR: Menu %s doesn't exist in $desc." $menu_name_untranslated]
	    return 1
	}
		
	set i [GiDMenu::_FindOptionIndex $i [lrange $option_name_untranslated 0 end-1] $desc]
	if { $i == -1 } {
	    return 1
	}
	if { ![info exists MenuEntries($i)] } {
	    set MenuEntries($i) ""
	}
	set tail_opt_name [$translationfunc [lindex $option_name_untranslated end]]
	if { $position == "end" } {
	    set position [expr {[llength $MenuEntries($i)]-1}]
	}
	if { $ins_repl == "insertafter" } {
	    incr position
	}
	if {[llength $MenuEntries($i)]<= $position} {
	    lappend MenuEntries($i) $tail_opt_name
	    lappend MenuCommands($i) "-np- $command"
	    if {[info exists MenuAcceler($i)]} {
		lappend MenuAcceler($i) $acceler
	    }
	    if {[info exists MenuIcons($i)]} {
		lappend MenuIcons($i) $icon
	    }
	} else {
	    if { $ins_repl == "replace" } {
		del_submenus $i $position $desc
		set MenuEntries($i) [lreplace $MenuEntries($i) $position $position $tail_opt_name]
		set MenuCommands($i) [lreplace $MenuCommands($i) $position $position "-np- $command"]
		if {[info exists MenuAcceler($i)] && [llength $MenuAcceler($i)] > $position} {
		    set MenuAcceler($i) [lreplace $MenuAcceler($i) $position $position $acceler]
		}
		if {[info exists MenuIcons($i)] && [llength $MenuIcons($i)] > $position } {
		    set MenuIcons($i) [lreplace $MenuIcons($i) $position $position $icon]
		}
	    } else {
		# "insert"
		set MenuEntries($i) [linsert $MenuEntries($i) $position  $tail_opt_name]
		set MenuCommands($i) [linsert $MenuCommands($i) $position  "-np- $command"]
		if {[info exists MenuAcceler($i)]} {
		    set MenuAcceler($i) [linsert $MenuAcceler($i) $position $acceler]
		}
		if {[info exists MenuIcons($i)]} {
		    set MenuIcons($i) [linsert $MenuIcons($i) $position $icon]
		}
		move_submenus $i $position $desc "down"
	    }

	}
    }
    return 0
}


# -----------------------------------------------------------------------------
#  GiDMenu::RemoveOption
#
#  Removes an option of a given menu.
#  menu_name_untranslated: the toplevel menu name untranslated
#  option_name_untranslated: the option name, can be a list for sublevels
#  e.g [list "Create#C#menu" "Arc#C#menu"]
#  it is also accepted the old name without comments: [list "Create" "Arc"]
#  prepost: PRE POST or PREPOST
#  translationfunc can be _ (the default value,for GiD standard strings) or = (for module strings)
#
#  Remember call GiDMenu::UpdateMenus to apply all changes.
# -----------------------------------------------------------------------------
proc GiDMenu::RemoveOption {menu_name_untranslated option_name_untranslated prepost {translationfunc _}} {
    if {$prepost!="PRE" && $prepost!="POST" && $prepost!="PREPOST"} {
	WarnWin [_ "Wrong arg. Must be PRE, POST or PREPOST."]
	return 1
    }

    if {$prepost eq "PRE"} {
	set where [list "" "PRE"]
    } elseif {$prepost eq "POST"} {
	set where [list "P" "POST"]
    } else {
	set where [list "" "PRE" "P" "POST"]
    }
    foreach {c desc} $where {
	upvar \#0 MenuNames$c MenuNames
	upvar \#0 MenuEntries$c MenuEntries
	upvar \#0 MenuCommands$c MenuCommands
	upvar \#0 MenuAcceler$c MenuAcceler
	upvar \#0 MenuIcons$c MenuIcons
	
	set i [GiDMenu::_FindIndex $menu_name_untranslated $desc $translationfunc]
	if { $i == "-1" } {
	    WarnWin [_ "ERROR: Menu %s doesn't exist in $desc." $menu_name]
	    return 1
	}
	set index [GiDMenu::_FindOptionIndex $i $option_name_untranslated $desc]
	if { $index == -1 } {
	    return 1
	}
	set i [join [lrange [split $index ,] 0 end-1] ,]
	set position [lindex [split $index ,] end]
	del_submenus $i $position $desc
	move_submenus $i $position $desc "up"
	set MenuEntries($i)  [lreplace $MenuEntries($i) $position $position ]
	set MenuCommands($i) [lreplace $MenuCommands($i) $position $position]
	if {[info exists MenuAcceler($i)] && [llength $MenuAcceler($i)] > $position} {
	    set MenuAcceler($i) [lreplace $MenuAcceler($i) $position $position ]
	}
	if {[info exists MenuIcons($i)] && [llength $MenuIcons($i)] > $position} {
	    set MenuIcons($i) [lreplace $MenuIcons($i) $position $position]
	}
    }
    return 0
}

# -----------------------------------------------------------------------------
#  GiDMenu::ModifyOption
#
#  Modify a previous menu identified by menu_name_untranslated option_name_untranslated and prepost
#  menu_name_untranslated: the toplevel menu name untranslated
#  option_name_untranslated: the option name, can be a list for sublevels
#  e.g [list "Create#C#menu" "Arc#C#menu"]
#  it is also accepted the old name without comments: [list "Create" "Arc"]
#  Arg prepost: PRE POST or PREPOST
#  Arg new_option_name new option name (-default- to keep the current)
#  Arg new_command is an optional new command (-default- to keep the current)
#  Arg new_acceler is an optional key accelerator, like "Control-s" (-default- to keep the current)
#  Arg new_icon is the name of an small icon (16x16 pixels) (-default- to keep the current)
#  Arg translationfunc can be _ (the default value,for GiD standard strings) or = (for module strings)
#
#  Remember call GiDMenu::UpdateMenus to apply all changes.
# -----------------------------------------------------------------------------
proc GiDMenu::ModifyOption { menu_name_untranslated option_name_untranslated prepost new_option_name \
    {new_command -default-} {new_acceler -default-} {new_icon -default-} {translationfunc _} } {

    if {$prepost!="PRE" && $prepost!="POST" && $prepost!="PREPOST"} {
	return -code error [_ "Wrong arg. Must be PRE, POST or PREPOST."]
    }
	
    if {$prepost eq "PRE"} {
	set where [list "" "PRE"]
    } elseif {$prepost eq "POST"} {
	set where [list "P" "POST"]
    } else {
	set where [list "" "PRE" "P" "POST"]
    }
    foreach {c desc} $where {
	upvar \#0 MenuNames$c MenuNames
	upvar \#0 MenuEntries$c MenuEntries
	upvar \#0 MenuCommands$c MenuCommands
	upvar \#0 MenuAcceler$c MenuAcceler
	upvar \#0 MenuIcons$c MenuIcons
	
	set i [GiDMenu::_FindIndex $menu_name_untranslated $desc $translationfunc]
	if { $i == "-1" } {
	    ::GidUtils::SetWarnLine [_ "ERROR: Menu %s doesn't exist in $desc." $menu_name]
	    return 1
	}
	
	set index [GiDMenu::_FindOptionIndex $i $option_name_untranslated $desc]
	if { $index == -1 } {
	    return 1
	}
	set i [join [lrange [split $index ,] 0 end-1] ,]
	set position [lindex [split $index ,] end]

	if { $new_option_name != "-default-" } {
	    set MenuEntries($i) [lreplace $MenuEntries($i) $position $position $new_option_name]
	}
	if { $new_command != "-default-" } {
	    set MenuCommands($i) [lreplace $MenuCommands($i) $position $position "-np- $new_command"]
	}
	if {[info exists MenuAcceler($i)] && [llength $MenuAcceler($i)] > $position} {
	    if { $new_acceler != "-default-" } {
		set MenuAcceler($i) [lreplace $MenuAcceler($i) $position $position $new_acceler]
	    }
	}
	if {[info exists MenuIcons($i)] && [llength $MenuIcons($i)] > $position } {
	    if { $new_icon != "-default-" } {
		set MenuIcons($i) [lreplace $MenuIcons($i) $position $position $new_icon]
	    }
	}
    }
    return 0
}


# -----------------------------------------------------------------------------
#  GiDMenu::_FindIndex
#
#  return the menu index
#  menu_name_untranslated (untranslated english menu name, for example "Files#C#menu"
#  prepost: PRE or POST (where to find)
#  translationfunc can be _ (for GiD standard strings) or = (for additional module strings)
# -----------------------------------------------------------------------------
proc GiDMenu::_FindIndex { menu_name_untranslated prepost {translationfunc _} } {
    global MenuNames MenuNamesP
    if {$prepost eq "PRE" } {
	upvar \#0 MenuNames TranslatedNames
    } else { ;#POST
	upvar \#0 MenuNamesP TranslatedNames
    }
    set i [lsearch $TranslatedNames [$translationfunc $menu_name_untranslated]]
    if { $i!= -1 } {
	return $i
    }
    if {$prepost eq "PRE" } {
	#back compatibility: Meshing menu renamed to Mesh in 7.5.9b
	if { $menu_name_untranslated == "Meshing" } {
	    set i [lsearch $TranslatedNames [$translationfunc "Mesh#C#menu"]]
	    if { $i != "-1" } {
		return $i
	    }
	}
    }
    #back compatibility: menus renamed from oldname to oldname#C#menu in 8.0
    set postfix "#C#menu"
    set i [lsearch $TranslatedNames [$translationfunc $menu_name_untranslated$postfix]]
    if { $i != "-1" } {
	return $i
    }
    return -1
}


# -----------------------------------------------------------------------------
#  return the submenu option index
#  parentindex: the index (from 0) of the parent menu
#  option_name_untranslated (untranslated english menu name, for example "Export#C#menu>IGES#C#menu")
#  prepost: PRE or POST (where to find)
#  insert:
#  translationfunc can be _ (for GiD standard strings) or = (for additional module strings)
# -----------------------------------------------------------------------------
proc GiDMenu::_FindOptionIndex { parentindex nameslist prepost {translationfunc _}} {
    global MenuEntries MenuEntriesP
    if {$prepost eq "PRE" } {
	upvar \#0 MenuEntries TranslatedEntries
    } else { ;#POST
	upvar \#0 MenuEntriesP TranslatedEntries
    }

    set postfix "#C#menu"
    set index $parentindex
    foreach suboption $nameslist {
	set found 0
	set cont 0
	if { ![info exists TranslatedEntries($index)] } {
	    return -1
	}
	foreach item $TranslatedEntries($index) {
	    if { [cmpoption $item [$translationfunc $suboption]] == 0 || \
		[cmpoption $item [$translationfunc $suboption$postfix]] == 0 } {
		append index ,$cont
		set found 1
		break
	    }
	    incr cont
	}
	if { !$found } {
	    return -1
	}
    }
    return $index
}

proc GiDMenu::cmpoption { a b } {
    if { [string match *... $a] } {
	set a [string range $a 0 end-3]
    }
    if { [string match *... $b] } {
	set b [string range $b 0 end-3]
    }
    return [string compare $a $b]
}


proc GiDMenu::del_submenus {i position pre} {
    global MenuNames MenuEntries MenuCommands MenuAcceler MenuIcons
    global MenuNamesP MenuEntriesP MenuCommandsP MenuAccelerP MenuIconsP

    if {$pre=="PRE"} {
	if {[info exists MenuEntries($i,$position)]} {
	    append i ",$position"
	    rec_del_menu $i MenuEntries MenuCommands MenuAcceler MenuIcons
	}
    }

    if {$pre=="POST"} {
	if {[info exists MenuEntriesP($i,$position)]} {
	    append i ",$position"
	    rec_del_menu $i MenuEntriesP MenuCommandsP MenuAccelerP MenuIconsP
	}
    }
}

proc GiDMenu::rec_del_menu {param ent com acc icons} {
    upvar $ent Entries
    upvar $com Commands
    upvar $acc Acceler
    upvar $icons Icons

    set m_length [llength $Entries($param)]

    set pos 0

    unset Entries($param)
    if {[info exists Commands($param)]} {unset Commands($param) }
    if {[info exists Acceler($param)]}  {unset Acceler($param) }
    if {[info exists Icons($param)]}  {unset Icons($param) }
	
    set base $param
    while {$pos<$m_length} {
	set param "$base,$pos"
	if {[info exists Entries($param)]} {rec_del_menu $param Entries Commands Acceler Icons}
	set pos [expr $pos +1]
    }

}

proc GiDMenu::move_submenus {i position pre inc} {
    global MenuNames MenuEntries MenuCommands MenuAcceler MenuIcons
    global MenuNamesP MenuEntriesP MenuCommandsP MenuAccelerP MenuIconsP
	
    set last_menu [llength [split $i ,]]
    incr last_menu

    if {$pre=="PRE"} {
	set j 1
	set m_length [llength $MenuEntries($i)]
	set base $i
	while {$position<$m_length} {
	
	    # read from bottom to top when insert, and opposite when removing
	    set param1 "$base,$position"
	    set param2 "$base,[expr $m_length - $j]"
	    if {[info exists MenuEntries($param1)]} {
		if {$inc=="up"}   {rec_move_menu $param1 $last_menu MenuEntries MenuCommands MenuAcceler MenuIcons $inc}
	    }
	
	    if {[info exists MenuEntries($param2)]} {
		if {$inc=="down"} {rec_move_menu $param2 $last_menu MenuEntries MenuCommands MenuAcceler MenuIcons $inc}
	    }
	    set position [expr $position +1]
	    set j [expr $j +1]
	}
    }


    if {$pre=="post"} {
	set m_length [llength $MenuEntriesP($i)]
	set base $i
	while {$position<$m_length} {
	    set param "$base,$position"
	    if {[info exists MenuEntriesP($param)]} {
		rec_move_menu $param $last_menu MenuEntriesP MenuCommandsP MenuAccelerP MenuIconsP $inc
	    }
	    set position [expr $position +1]
	}
    }
}


proc GiDMenu::dec_inc_menu {param last_menu inc} {
    set up_menus   ""
    set down_menus ""
    set menu ""
    set j 0
    set num 0

    while {$num < [expr $last_menu-1] && $j<[string length $param]} {
	set char [string index $param $j]
	append up_menus $char
	if {$char == ","} {set num [expr $num +1]}
	set j [expr $j +1]
    }

    set char [string index $param $j]
    while {$char != "," && $j<[string length $param]}  {
	append menu $char
	set j [expr $j +1]
	set char [string index $param $j]
    }

    while {$j <= [string length $param]} {
	append down_menus $char
	set j [expr $j +1]
	set char [string index $param $j]
    }
    append down_menus $char

    if {$inc=="down"} {
	set menu [expr $menu + 1]
    } else {
	set menu [expr $menu - 1]
    }
    set new_menu $up_menus$menu$down_menus
    return $new_menu
}


proc GiDMenu::rec_move_menu {param last_menu ent com acc icons inc} {
    upvar $ent Entries
    upvar $com Commands
    upvar $acc Acceler
    upvar $icons Icons

    set m_length [llength $Entries($param)]
    set pos 0

    set new_menu [dec_inc_menu $param $last_menu $inc]

    set Entries($new_menu) $Entries($param)
    unset Entries($param)

    if {[info exists Commands($param)]} {
	set Commands($new_menu) $Commands($param)
	unset Commands($param)
    }
    if {[info exists Acceler($param)]}  {
	set Acceler($new_menu)  $Acceler($param)
	unset Acceler($param)
    }

    if {[info exists Icons($param)]}  {
	set Icons($new_menu)  $Icons($param)
	unset Icons($param)
    }
    set base $param
    while {$pos<$m_length} {
	set param "$base,$pos"
	if {[info exists Entries($param)]} {rec_move_menu $param $last_menu Entries Commands Acceler Icons $inc}
	set pos [expr $pos +1]
    }
}


#utilities not interacting with GiD

namespace eval MathUtils {
    variable Epsilon 1e-10
    variable Pi [ expr 2.0 * acos(0.0)]
}

#pi is a list with x y z? coordinates, return the diff vector p2-p1 as a list
proc MathUtils::VectorDiff { p1 p2 } {
    set diff ""
    foreach i $p1 j $p2 {
	lappend diff [expr $j-$i]
    }
    return $diff
}

proc MathUtils::VectorSum { p1 p2 } {
    set sum ""
    foreach i $p1 j $p2 {
	lappend sum [expr {$j+$i}]
    }
    return $sum
}

#pi is a list with x y z? coordinates, return the dot prod value
proc MathUtils::VectorDotProd { p1 p2 } {
    set dotprod 0
    foreach i $p1 j $p2 {
	set dotprod [expr $dotprod+($i*$j)]
    }
    return $dotprod
}

#pi is a list with x y z coordinates, return the vectorial prod vector as list
proc MathUtils::VectorVectorialProd { p1 p2 } {
    set ax [lindex $p1 0]
    set ay [lindex $p1 1]
    set az [lindex $p1 2]
    set bx [lindex $p2 0]
    set by [lindex $p2 1]
    set bz [lindex $p2 2]
    set vectorprod [expr $ay*$bz-$az*$by]
    lappend vectorprod [expr $az*$bx-$ax*$bz]
    lappend vectorprod [expr $ax*$by-$ay*$bx]
    return $vectorprod
}

#multiplication of a scalar 'a' by a vector 'p'
proc MathUtils::ScalarByVectorProd { a p } {
    foreach i $p {
	lappend sum [expr {$a*$i}]
    }
    return $sum
}

#p is a list with x y z? coordinates, return a unitary vector
proc MathUtils::VectorNormalized { p } {
    variable Epsilon
    set m [::MathUtils::VectorModulus $p]
    if { [expr abs($m)] <  $Epsilon } {
	return $p
    }
    return [::MathUtils::ScalarByVectorProd [expr 1.0/$m] $p]
}

#p is a list with x y z? coordinates, return the modulus value
proc MathUtils::VectorModulus { p } {
    return [expr sqrt([MathUtils::VectorDotProd $p $p])]
}

#pi is a list with x y z? coordinates, return the distance value
proc MathUtils::VectorDistance { p1 p2 } {
    return [MathUtils::VectorModulus [MathUtils::VectorDiff $p1 $p2]]
}

#pi is a list with x y z? coordinates, return the distance^2 value
#(for some uses must be faster to compare dist^2 avoiding calculate the square root)
proc MathUtils::VectorDistance2 { p1 p2 } {
    set p [MathUtils::VectorDiff $p1 $p2]
    return [MathUtils::VectorDotProd $p $p]
}

#pi is a list with x y z? coordinates, return the angle value in radians
proc MathUtils::VectorAngle { pvertex p2 p3} {
    set v1 [MathUtils::VectorDiff $p2 $pvertex]
    set modul1 [MathUtils::VectorModulus $v1]
    if { $modul1<$MathUtils::Epsilon } return ""

    set v2 [MathUtils::VectorDiff $p3 $pvertex]
    set modul2 [MathUtils::VectorModulus $v2]
    if { $modul2<$MathUtils::Epsilon } return ""

    set cosangle [expr [MathUtils::VectorDotProd $v1 $v2]/($modul1*$modul2)]
    set angle [expr acos($cosangle)]
    return $angle
}

proc MathUtils::FromRadToDegree { x } {
    return [expr $x*180.0/$MathUtils::Pi]
}

proc MathUtils::FromDegreeToRad { x } {
    return [expr $x*$MathUtils::Pi/180.0]
}



#utilities interacting with GiD
namespace eval GidUtils {
}

#register a tcl command for the GiD Undo/Batch, and eval this command
proc GidUtils::EvalAndEnterInBatch { args } {
    GiD_EnterInBatch "\"-tcl- $args\"\n"
    eval $args
}

# returns the distance (or null) between two selected coordinates
proc GidUtils::GetDistance { } {
    set p1 [GidUtils::GetCoordinates [_ "Enter first point (ESC to leave)"]]
    if { $p1=="" } return ""
    set p2 [GidUtils::GetCoordinates [_ "Enter second point (ESC to leave)"]]
    if { $p2=="" } return ""

    set d [MathUtils::VectorDistance $p1 $p2]

    GidUtils::SetWarnLine [_ "Distance=%s" $d]
    return $d
}

proc GidUtils::GetVector { } {
    set p1 [GidUtils::GetCoordinates [_ "Enter origin point (ESC to leave)"]]
    if { $p1=="" } return ""
    set p2 [GidUtils::GetCoordinates [_ "Enter destination point (ESC to leave)"]]
    if { $p2=="" } return ""

    set p [MathUtils::VectorDiff $p1 $p2]

    GidUtils::SetWarnLine [_ "Vector=%s" $p]
    return $p
}

#return the angle in degrees
proc GidUtils::GetAngle { } {
    set p1 [GidUtils::GetCoordinates [_ "Enter vertex point (ESC to leave)"]]
    if { $p1=="" } return ""
    set p2 [GidUtils::GetCoordinates [_ "Enter second point (ESC to leave)"]]
    if { $p2=="" } return ""
    set p3 [GidUtils::GetCoordinates [_ "Enter third point (ESC to leave)"]]
    if { $p3=="" } return ""

    set angle [MathUtils::FromRadToDegree [MathUtils::VectorAngle $p1 $p2 $p3]]
    GidUtils::SetWarnLine [_ "Angle=%s" $angle]
    return $angle
}

#return the curvature o the curve with id=num relative to the parameter t (from 0.0 to1.0)
proc GidUtils::GetCurveCurvature { num t } {
    set dert [GiD_Info parametric line $num deriv_t $t]
    set dertt [GiD_Info parametric line $num deriv_tt $t]
    set a [::MathUtils::VectorModulus [MathUtils::VectorVectorialProd $dert $dertt]]
    set b [::MathUtils::VectorModulus $dert]
    return [expr $a/($b*$b*$b)]
}


# It returns the enties identifier numbers (for example 1 3 5:8 represents 1 3 5 6 7 8)
# type can be "Points" "Lines" "Surfaces" "Volumes" "Dimensions" "AllTypes" "Nodes" or "Elements"
# selectmode can be "single" or "multiple"
# for AllTypes selection returns some as
# Points 19 23 Lines 28:29 32:33 37 Surfaces 1 6 21:22 Volumes 1 Dimension 1:2
# or for mesh entities some as Nodes 236 244 Elements 4122 4130 4152 4165
proc GidUtils::PickEntities { type selectmode } {
    gidmakecurrent
    set wait 1
    switch $selectmode {
      single {
	set args_sel "justone $type"
      }
      multiple {
	set args_sel "nojustone $type"
      }
      default {
	set args_sel "$type $selectmode escape"
	set wait 0
      }
    }
    GiD_Process Mescape selectentities {*}$args_sel
    if { $wait } {
      vwait ::GidPriv(selection)
    }
    if { $selectmode == "single" } {
	set selection [lindex {*}$::GidPriv(selection) 1]
    } else {
	set selection [lrange {*}$::GidPriv(selection) 1 end]
    }
    return $selection
}


# returns the coordinates x y z Or null selecting from the screen
# GidPriv(selcoord,id) store the point id (-1 nothing selected, 0 new point, >0 point id)
# to get the number of a existent point, cal also use PickEntities
# PickMode can be "Join" (default)    : to pick existing points
#                 "NoJoin"            : to pick a new point generic
#                 "PointInLine"       : to pick a new point on a curve
#                 "PointInSurface"    : to pick a new point on a curve
#                 "TangentInLine"     : to pick a new unitary curve tangent vector
#                 "NormalInSurface"   : to pick a new unitary surface normal vector
#                 "ArcCenter"         : to pick a new point center of an arc
#                 "LineParameter"     : to pick a new point in a line, specifying the t parameter
#                 "SurfaceParameters" : to pick a new point in a surface, specifying the u,v parameters
#                 "MapPointSurface"   : to pick a new point in a surface, specifying a near initial point

proc GidUtils::GetCoordinates { { WarnLineMsg "" } { PickMode  "Join" } } {
    if { $WarnLineMsg  == "" } { set WarnLineMsg [_ "Enter point (ESC to leave)"] }
    gidmakecurrent
    GidUtils::SetWarnLine $WarnLineMsg
    GidUtils::DisableWarnLine
    GiD_Process escape escape escape escape utilities id
    update idletasks
    if { [lsearch -exact [lrange $::ActionsMenuVar 0 end-1] $PickMode] != -1 } {
	GiD_Process $PickMode
    }
    GidUtils::EnableWarnLine

    vwait ::GidPriv(selcoord,id)
    if { $::GidPriv(selcoord,id) != -1 } {
      set xyz "$::GidPriv(selcoord,x) $::GidPriv(selcoord,y) $::GidPriv(selcoord,z)"
    } else {
      set xyz ""
    }
    GidUtils::DisableWarnLine
    GiD_Process escape escape
    GidUtils::EnableWarnLine
    return $xyz
}


#to compare if a GiD version is more recent of another
proc GidUtils::VersionNumber { VersionString } {
    set patron {([0-9]+)[.]([0-9]+)[.]?([0-9]*)([a-z]?)(?:-beta([0-9]+))?}
    regexp -nocase $patron $VersionString total v1 v2 v3 letra betanumber
    set res 0
    if { [info exists v1] } {
	if { $v1 != "" } { set res [expr $res+$v1*100000] }
    }
    if { [info exists v2] } {
	if { $v2 != "" } { set res [expr $res+$v2*1000] }
    }
    if { [info exists v3] } {
	if { $v3 != "" } { set res [expr $res+$v3] }
    }
    return $res
}

proc GidUtils::VersionSubchar { VersionString } {
    set patron {([0-9]+)[.]([0-9]+)[.]?([0-9]*)([a-z]?)(?:-beta([0-9]+))?}
    regexp -nocase $patron $VersionString total v1 v2 v3 letra betanumber
    set res ""
    if { [info exists letra] } { set res $letra }
    return $res
}

proc GidUtils::VersionBetaNumber { VersionString } {
    set patron {([0-9]+)[.]([0-9]+)[.]?([0-9]*)([a-z]?)(?:-beta([0-9]+))?}
    regexp -nocase $patron $VersionString total v1 v2 v3 letra betanumber
    set res ""
    if { [info exists betanumber] } { set res $betanumber }
    return $res
}

# this function always returns -1 if GiD is before to the
# 6.2.0 version, first including info GiDVersion
# return -1 if GidPriv(Version) < Version
# return  0 if GidPriv(Version) == Version
# return  1 if GidPriv(Version) > Version

#sample use for a probletype that need GiD7.4b or later new features:
#    set VersionRequired "7.4b"
#    set comp -1
#    catch {
#      set comp [::GidUtils::VersionCmp $VersionRequired]
#    }
#    if { $comp < 0 } {
#        WarnWin "Error: This interface requires GiD $VersionRequired or later."
#    }

proc GidUtils::VersionCmp { Version } {
    if { [catch { set GidVersion [GiD_Info GiDVersion] }] }  {
	return -1
    }
    if { $GidVersion == "Unknown" } {
	return -1
    }
    set Number1 [GidUtils::VersionNumber $GidVersion]
    set Number2 [GidUtils::VersionNumber $Version]
    if { $Number1 < $Number2 } { return -1 }
    if { $Number1 > $Number2 } { return 1 }
    set c1 [GidUtils::VersionSubchar $GidVersion]
    set c2 [GidUtils::VersionSubchar $Version]
    if { $c1 < $c2 } { return -1 }
    if { $c1 > $c2 } { return 1 }
    #assumed more recent the non-beta version
    set BetaNumber1 [GidUtils::VersionBetaNumber $GidVersion]
    set BetaNumber2 [GidUtils::VersionBetaNumber $Version]
    if { $BetaNumber1 != "" } {
	if { $BetaNumber2 != "" } {
	    if { $BetaNumber1 < $BetaNumber2 } { return -1 }
	    if { $BetaNumber1 > $BetaNumber2 } { return 1 }
	} else {
	    return -1
	}
    } else {
	if { $BetaNumber2 != "" } { return 1 }
    }

    return 0
}

#to access to values by fieldname instead index (more easy to maintain code)
proc GidUtils::GetProblemDataValue { fieldname } {

    foreach "foundname foundvalue" [list "" ""] break
    foreach "name value" [lrange [GiD_Info gendata] 1 end] {
	if { [string equal $fieldname $name] } { return $value }
	if { [string match $fieldname* $name] } {
	    if { $foundname != "" } {
		error "error: '$fieldname' has more than one names: $foundname and $name"
	    }
	    foreach "foundname foundvalue" [list $name $value] break
	}
    }
    if { $foundname == "" } {
	error "error: '$fieldname' is not found in Problem data"
    }
    return $foundvalue
}

#to access to values by fieldname instead index (more easy to maintain code)
proc GidUtils::GetCurrentIntervalDataValue { fieldname } {

    foreach "foundname foundvalue" [list "" ""] break
    foreach "name value" [lrange [GiD_Info intvdata] 1 end] {
	if { [string equal $fieldname $name] } { return $value }
	if { [string match $fieldname* $name] } {
	    if { $foundname != "" } {
		error "error: '$fieldname' has more than one names: $foundname and $name"
	    }
	    foreach "foundname foundvalue" [list $name $value] break
	}
    }
    if { $foundname == "" } {
	error "error: '$fieldname' is not found in Problem data"
    }
    return $foundvalue
}

#change status of redraws, etc to speed or enhance scripts

#after use a pair: GidUtils::DisableGraphics - GidUtils::EnableGraphics
#the previous stated is automatically restored (using a stack)
proc GidUtils::DisableGraphics { } {
    variable state
    foreach i [list windows graphinput graphics warnline] {
	lappend state($i) [.central.s disable $i]
	.central.s disable $i 1
    }
}

proc GidUtils::EnableGraphics { } {
    variable state
    foreach i [list windows graphinput graphics warnline] {
	if { [info exists state($i)] && [llength $state($i)] } {
	    .central.s disable $i [lindex $state($i) end]
	    set state($i) [lrange $state($i) 0 end-1]
	} else {
	    .central.s disable $i 0
	}
    }
    return
}

# i must be: windows graphinput graphics warnline writebatch
proc GidUtils::Disable { i } {
    variable state
    if { [lsearch {windows graphinput graphics warnline writebatch} $i] == -1 } {
	return 1
    }
    lappend state($i) [.central.s disable $i]
    .central.s disable $i 1
    return 0
}

# i must be: windows graphinput graphics warnline writebatch
proc GidUtils::Enable { i } {
    variable state
    if { [lsearch {windows graphinput graphics warnline writebatch} $i] == -1 } {
	return 1
    }
    if { [info exists state($i)] && [llength $state($i)] } {
	set prevstate [lindex $state($i) end]
	.central.s disable $i [lindex $state($i) end]
	set state($i) [lrange $state($i) 0 end-1]
    } else {
	.central.s disable $i 1
    }
    return 0
}

proc GidUtils::DisableWarnLine {  } {
    return [GidUtils::Disable warnline]
}

proc GidUtils::EnableWarnLine {  } {
    return [GidUtils::Enable warnline]
}


proc GidUtils::WaitState { w } {
    $w conf -cursor watch
    .central.s waitstate 1
    update
}

proc GidUtils::EndWaitState { w } {
    .central.s waitstate 0
    $w conf -cursor ""
    update
}

proc GidUtils::SetWarnLine { text } {
    if { [.central.s disable warnline] } return
    ChangeWarnLineVar $text
}

#Utility to show a progress bar.
#varname must be a global variable name with the current value to be showed automatically
#stopproc is an optional callback procedure name to be called when the user press stop
#when the current value is equal to max the advance bar is automatically deleted
#Note: only is possible to use one progress bar
#example:
#  global MyVarName
#  set MyVarName 0
#  ::GidUtils::CreateAdvanceBar [= "Loading file"] [= "Percentage"] ::MyVarName 100  [list WarnWin [= "User stop"]]
#  # ...
#  set MyVarName 20

proc GidUtils::CreateAdvanceBar { title text varname { max 100 } { stopproc "" } } {
    variable ab_maximum
    variable ab_varname
    variable ab_stopproc
    set ab_maximum $max
    set ab_varname $varname
    set ab_stopproc $stopproc
    set ::GidPriv(StopLoading) 0
    AdvanceBar 0 $ab_maximum 0 $title $text
    if { $ab_varname != "" } {
	trace add variable $ab_varname write "::GidUtils::UpdateAdvanceBar ;#"
    }
    if { $ab_stopproc != "" } {
	trace add variable ::GidPriv(StopLoading) write "::GidUtils::CallStopProc ;#"
    }
}

proc GidUtils::UpdateAdvanceBar { } {
    variable ab_maximum
    variable ab_varname
    if { ![info exists $ab_varname] } return
    if { [set $ab_varname] >= $ab_maximum } {
	::GidUtils::DeleteAdvanceBar
    } else {
	AdvanceBar [set $ab_varname] $ab_maximum [set $ab_varname]
    }
}

proc GidUtils::CallStopProc { } {
    variable ab_stopproc
    if { $::GidPriv(StopLoading) == "0" } return
    ::GidUtils::DeleteAdvanceBar
    if { $ab_stopproc != "" } {
	eval $ab_stopproc
    }
}

# proc GidUtils::UpdateAdvanceBar { current } {
#     variable ab_maximum
#     AdvanceBar $current $ab_maximum $current
# }

proc GidUtils::DeleteAdvanceBar { } {
    variable ab_maximum
    variable ab_varname
    variable ab_stopproc
    if { $ab_varname != "" } {
	trace remove variable $ab_varname write "GidUtils::UpdateAdvanceBar ;#"
    }
    if { $ab_stopproc != "" } {
	trace remove variable ::GidPriv(StopLoading) write "::GidUtils::CallStopProc ;#"
    }
    AdvanceBar $ab_maximum $ab_maximum $ab_maximum
}

#to show a splash image, for example:
#global GIDDEFAULT
#GidUtils::Splash [file join $GIDDEFAULT bitmaps about.gif]
#VersionAndPos: list with three items: version_label posx posy
proc GidUtils::Splash { imagefile {w .splash} {selfclose 1} {VersionAndPos ""} } {
    if { [GidUtils::AreWindowsDisabled] } return
    #if selfclose==0 ignore the SplashWindow GiD variable and show the window
    #tipically selfclose==1 for splash initial windows (then use SplashWindow)
    #          selfclose==0 for help about windows (then show avoiding the SplashWindow state)
    if { ![GiD_Set SplashWindow] && $selfclose==1 } return

    if { ![file exists $imagefile] } {
	WarnWinText [_ "Not exists file '%s'" $imagefile]
	return
    }
    if { [ winfo exists $w]} {
	destroy $w
	update
    }
    toplevel $w
    set im [image create photo -file $imagefile]
    set x [expr [winfo screenwidth $w]/2-[image width $im]/2]
    set y [expr [winfo screenheight $w]/2-[image height $im]/2]
    wm geom $w +$x+$y
    wm transient $w .gid
    wm overrideredirect $w 1
    pack [label $w.l -image $im -relief ridge -bd 2]

    if { $VersionAndPos != "" } {
	if { [llength $VersionAndPos] == 3 } {
	    set VersionNumber [lindex $VersionAndPos 0]
	    set x [lindex $VersionAndPos 1]
	    set y [lindex $VersionAndPos 2]
	} else {
	    set VersionNumber [lindex $VersionAndPos 0]
	    set x 62
	    set y 62
	}
	label .splash.lv -text $VersionNumber -foreground white -relief raised -bd 3 \
	    -background blue4 -font "verdana 8"
	place .splash.lv -x $x -y $y
    }

    bind $w <1> "destroy $w"
    bind $w <KeyPress> "destroy $w"
    raise $w .gid
    grab $w
    focus $w
    update
    if { $selfclose } {
	after 2000 "if { [ winfo exists $w] } { destroy $w }"
    }
}

proc GidUtils::GetCurrentLanguage { } {
    #instead for example "es_es_modern", return only "es"
    return [string range [::msgcat::mclocale] 0 1]
}

proc GidUtils::SetCurrentLanguage { newlan {updatewidgets 1 } } {
    global GidPriv
    if { ![info exists GidPriv(Language)] || $GidPriv(Language) != $newlan } {
	set GidPriv(Language) $newlan
    }
    if { [::GidUtils::GetCurrentLanguage] != $newlan } {
	MsgcatSetLocale $newlan
    }

    if { $updatewidgets } {
	UpdateWidgetsLanguage
    }
    #raise ChangedLanguage as a "event"
    if { [info procs ::ChangedLanguage] != "" } {
	::ChangedLanguage $newlan
    } else {
	catch {::ChangedLanguage $newlan}
    }
}

proc GidUtils::GetAllLanguages { } {
    return [MsgcatGetLocales]
}


#GiD selection accept and returns entities idetifiers (integers > 0) as
#a "compacted" list, the sintax "a:a+n" is interpreted as "a a+1 a+2 ... a+n"

#instead "4 8 9 10 11 15" compact list --> "4 8:11 15"
#items must be positive integer numbers (can be unordered)
proc GidUtils::CompactNumberList { a } {
    set ret ""
    set istart 0
    if { [string match "*:*" $a] } {
	#imput list is compacted
	return $a
    }
    foreach i [lsort -integer $a] {
	if { !$istart} {
	    set istart $i
	} else {
	    if {$i != [expr $iprev+1]} {
		if { $istart != $iprev } {
		    if { [expr $istart +1] == $iprev } {
		        append ret "$istart $iprev "
		    } else {
		        append ret "$istart:$iprev "
		    }
		} else {
		    append ret "$istart "
		}
		set istart $i
	    }
	}
	set iprev $i
    }
    if { $istart } {
	if { $istart != $iprev } {
	    if { [expr $istart +1] == $iprev } {
		append ret "$istart $iprev"
	    } else {
		append ret "$istart:$iprev"
	    }
	} else {
	    append ret "$istart"
	}
    }
    return $ret
}

#instead "4 8:11 15"  uncompact list --> "4 8 9 10 11 15"
#items must be positive increasing integer numbers
proc GidUtils::UnCompactNumberList { a } {
    set ret ""
    foreach item $a {
	if { [regexp   {([0-9]+)[ ]*:[ ]*([0-9]+)} $item dummy istart iend] } {
	    for { set i $istart } { $i <= $iend } { incr i } {
		lappend ret $i
	    }
	} else {
	    lappend ret $item
	}
    }
    return $ret
}

#function to close some window if it exists, window can be:
# CALCULATE COORDINATES COPY HELP LAYER LIST MESHQUALITY MOVE PAGESETUP
# PREFERENCES SELECTION TOOLBARS WARNING NOTES
proc GidUtils::CloseWindow { window } {
    set w ""
    switch $window {
	CALCULATE {
	    set w .gid.wrun
	}
	COORDINATES {
	    set w .gid.wSetCoord
	}
	COPY {
	    set w .gid.wcopymove
	}
	HELP {
	    set w .gid.helpref
	}
	LAYER {
	    set w .gid.central.wlay
	}
	LIST {
	    set w gid.listent
	}
	MESHQUALITY {
	    set w .gid.meshq
	}
	MOVE {
	    set w .gid.wcopymove
	}
	PAGESETUP {
	    set w .gid.wPageSetup
	}
	PREFERENCES {
	    set w .gid.wvar
	}
	SELECTION {
	    set w .gid.filter_win
	}
	TOOLBARS {
	    set w .gid.toolbars
	}
	WARNING {
	    set w .gid.warning
	}
	NOTES {
	    set w .gid.texteditor
	}
    }
    if { $w != "" } {
	if { [winfo exists $w] } {
	    if { $::LAYERS_TREE &&  $window == "LAYER" && [winfo class $w] == "Panedwindow" } {
		close_window_inside $w
	    } else {
		destroy $w
	    }
	}
    } else {
	WarnWin [_ "Unexpected window %s" $window]
    }
}

#returns a non existent filename inside the temporal directory
proc GidUtils::GetTmpFilename { {extension .tmp} } {
    global tcl_platform env
    set tmpdir ""
    if { $tcl_platform(platform) == "unix" } {
	set tmpdir /tmp
    } else {
	if { [info exists env(TMP)] } {
	    set tmpdir $env(TMP)
	} elseif { [info exists env(TEMP)] } {
	     set tmpdir $env(TEMP)
	} else {
	    set tmpdir "$env(windir)/Temp"
	}
    }
    set tmpname ""
    for { set i 0 } { $i < 10000 } {incr i} {
	set tmpname [file join $tmpdir gid$i]$extension
	if { ![file exists $tmpname] } break
    }
    return $tmpname
}

#return 0 if is not a link,  1 if the file is a link or  2 if it is Windows shell link
proc GidUtils::IsLink { lnk } {
    set islink 0
    if { ![file exists $lnk] } {
	return 0
    }
    if { [string match "windows" $::tcl_platform(platform)] } {
	set type [file type $lnk]
	if { $type == "file" && [string tolower [file extension $lnk]] == ".lnk" } {
	    set fp [open $lnk]
	    fconfigure $fp -encoding binary -translation binary -eofchar {}
	    binary scan [read $fp 4] i tmp
	    if { $tmp == "76" } {
		binary scan [read $fp 16] h32 tmp
		if { $tmp == "10412000000000000c00000000000064" } {
		    set islink 2 ;#windows shell link
		}
	    }
	    close $fp
	}
	if { !$islink } {
	    if { ![catch  {set res [file readlink $lnk]}] } {
		set islink 1
	    }
	}
    } else {
	if { ![catch  {set res [file readlink $lnk]}] } {
	    set islink 1
	}
    }
    return $islink
}

#return the name pointed by a link or Windows shell link
proc GidUtils::ReadLink { lnk } {
    set res ""
    set islink [IsLink $lnk]
    if { $islink == 1 } {
	catch {set res [file readlink $lnk]} err
    } elseif { $islink == 2 } {
#         set fp [open $lnk]
#         fconfigure $fp -encoding binary -translation binary -eofchar {}
#         foreach snip [split [read $fp] \x00] {
#             if {[regexp {^[A-Z]:\\} $snip] && [file exists $snip]} {
#                 lappend res $snip
#             }
#         }
#         close $fp
#         set res [join $res]
	catch { set res [GiD_GetInfoLnkFile $lnk] }
    }
    return $res
}

#return the relative (to basedir) path of target
#useful to have the relative path of a problemtype (relative to the \problemtypes GiD directory)
#e.g: GidUtils::MakeRelativePath {C:\gid\problemtypes} {C:\gid\problemtypes\ansys55\ansys55_3d}
#will return "ansys55\ansys55_3d"
proc GidUtils::MakeRelativePath { basedir target } {
    set bparts [file split [file normalize $basedir]]
    set tparts [file split [file normalize $target]]

    if {[lindex $bparts 0] eq [lindex $tparts 0]} {
	# If the first part doesn't match - there is no good relative path
	set blen [llength $bparts]
	set tlen [llength $tparts]
	for {set i 1} {$i < $blen && $i < $tlen} {incr i} {
	    if {[lindex $bparts $i] ne [lindex $tparts $i]} { break }
	}
	set path [lrange $tparts $i end]
	for {} {$i < $blen} {incr i} {
	    set path [linsert $path 0 ..]
	}
	set target [join $path [file separator]]
    }
    #file join to change \ by / also for windows
    return [file join $target]
}

proc GiveFileInsideProblemType { ptname ext {checkifexists 0} } {
    if { [string tolower [file extension $ptname]] == ".gid" } {
	set ptname [file rootname $ptname]
    }
    set tailname [file tail $ptname]
    set ptname $ptname.gid
    if { [file pathtype $ptname] == "relative" } {
	set fullptname [file join $::GIDDEFAULT problemtypes]
	foreach item [file split $ptname] {
	    set fullptname [file join $fullptname $item]
	    if { [GidUtils::IsLink $fullptname.lnk] == 2 } {
		set fullptname [GidUtils::ReadLink $fullptname.lnk]
	    }
	}
    } else {
	set fullptname $ptname
	if { [GidUtils::IsLink $fullptname.lnk] == 2 } {
	    set fullptname [GidUtils::ReadLink $fullptname.lnk]
	}
    }

    set filename [file join $fullptname $tailname$ext]
    if { $checkifexists } {
	if { ![file exists $filename] } {
	    set filename ""
	}
    }
    return $filename
}

#return the full path to a relative or absolute problemtype name, e.g.
#return "" if is not a valid GiD problemtype (must have some problemtype file .bas, .cnd, etc)
#in:  nastran3/nastran            out: C:/GiD project/problemtypes/nastran3/nastran
#in:  C:/tmp/nastran3/nastran     out: C:/tmp/nastran3/nastran
proc GiveProblemTypeFullname { ptname } {
    set checkifexists 1
    if { [GiveFileInsideProblemType $ptname ".geo" $checkifexists] != "" } {
	#it is a model, not a problemtype
	return ""
    }
    set fullfilename ""
    foreach ext {.bas .cnd .prb .mat .tcl .tbe .bat} {
	set fullfilename [GiveFileInsideProblemType $ptname $ext $checkifexists]
	if { $fullfilename != "" } {
	    break
	}
    }
    if { $fullfilename != "" } {
	set fullpath [file root [file dirname $fullfilename]]
    } else {
	set fullpath ""
    }
    return $fullpath
}

#create a raster image from a widget c with Img
# image create photo -format window -data $c

proc GidUtils::CaptureWindow { win } {
    # package require Img
    regexp {([0-9]*)x([0-9]*)\+([0-9]*)\+([0-9]*)} [winfo geometry $win] - w h x y
    # Make the base image based on the window
    set image [image create photo -format window -data $win]
    foreach child [winfo children $win] {
	CaptureWindowSub $child $image 0 0
    }
    #save image to disk
    #$image write $filename -format GIF
    return $image
}

proc GidUtils::CaptureWindowSub { win image px py } {
    if {![winfo ismapped $win]} {
	return
    }
    regexp {([0-9]*)x([0-9]*)\+([0-9]*)\+([0-9]*)} [winfo geometry $win] - w h x y
    incr px $x
    incr py $y
    # Make an image from this widget
    set tempImage [image create photo -format window -data $win]
    # Copy this image into place on the main image
    $image copy $tempImage -to $px $py
    image delete $tempImage
    foreach child [winfo children $win] {
	CaptureWindowSub $child $image $px $py
    }
}

#to print a integer number over an image (create a new image if dst == "")
proc GidUtils::PrintNumberOverImage { number {dst ""} {xorig 0} {yorig 0} } {
    if { ![string is integer $number] } {
	return $dst
    }
    #gif black digits 0 to 9 with transparent background
    set src [image create photo -data {
	    R0lGODlhRgAIAJEAANnZ2QAAAP///////yH5BAEAAAAALAAAAABGAAgAAALM
	    RBIEO84igmBExF0QjIiIC0oAQLEB4AAISAQABMExiDCjABAAAQQEACCCYMRZ
	    XBDsuAABChAAEEAIAAIgCCF8C7Q4CgIAQfDj7uIoEARAAAEBCACgAABB8C0w
	    AikIAEQEBYCICIIdZxFBcAzCwoACAATBh8ALowBB8CEsLghGHMQFwYkICwMK
	    ABAEP+4sjGIT/LC4OAphx8VdEBCDMAuCHWdxQbDj4sIoAAQABMGOC4ujABAR
	    BDsOKDYAHMGPg4ggGBFxEAQ/Ig6CEgAUADs=
	}]
    set w 7
    set h 8
    set n [string length $number]
    if { $dst == "" } {
	set dst [image create photo -width [expr $n*$w] -height $h]
	$dst blank
    }
    for {set i 0} {$i < $n} {incr i} {
	set digit [string index $number $i]
	set x1 [expr $digit*$w]
	set x2 [expr $i*$w+$xorig]
	set y2 $yorig
	$dst copy $src -from $x1 0 [expr $x1+$w] $h -to $x2 $y2  [expr $x2+$w] [expr $y2+$h]
    }
    image delete $src
    return $dst
}

#hide the toolbar and avoid to be showed again from the Toolbar window
#after use a pair: GidUtils::DisableToolbar - GidUtils::EnableToolbar
#the previous stated is automatically restored (using a stack)
#name is the toolbar name: "Right buttons" "Command line" "Up menu"
# "Geometry & View bar" "Standard bar" "View results bar" "MacrosToolbar"
proc GidUtils::DisableToolbar { name } {
    variable state
    global ToolbarsPriv GidPriv
    set item ""
    foreach i $ToolbarsPriv(toolbars) {
	if { $name == [lindex $i 0] } {
	    set item $i
	    break
	}
    }
    if { $item == "" } return
    if { ![info exists GidPriv([lindex $item 1])] } {
	set GidPriv([lindex $item 1]) [list NONE {} [lindex $item 2]]
    }
    lappend state(toolbar,$name) $GidPriv([lindex $item 1])

    [lindex $item 2] NONE
    set ToolbarsPriv(Hide,$name) 1
}

proc GidUtils::EnableToolbar { name } {
    variable state
    global ToolbarsPriv GidPriv
    set item ""
    foreach i $ToolbarsPriv(toolbars) {
	if { $name == [lindex $i 0] } {
	    set item $i
	    break
	}
    }
    if { $item == "" } return
    if { [info exists state(toolbar,$name)] && [llength $state(toolbar,$name)] } {
	set GidPriv([lindex $item 1]) [lindex $state(toolbar,$name) end]
	set state(toolbar,$name) [lrange $state(toolbar,$name) 0 end-1]
    } else {
	set GidPriv([lindex $item 1]) [list NONE {} [lindex $item 2]]
    }

    [lindex $item 2] [lindex $GidPriv([lindex $item 1]) 0]
    set ToolbarsPriv(Hide,$name) 0
}


#to visually show thousands separated by a character like ,
#e.g. change 12345 by 12,345
proc GidUtils::AddThousandsSeparator { str { sep ,}} {
    # let's hide whats right at the decimal point
    while { [ regsub {(.*\d\.(\d\d\d)+)(\d+.*)$} $str "\\1|!|!|!|\\3" str]} {
    }
    while { [ regsub {^(.*[-+]?\d+)(\d\d\d)} $str "\\1$sep\\2" str]} {
    }
    regsub -all {\|\!\|\!\|\!\|} $str {} str
    return $str
}

#to save a georeference file when saving an image
#if read this image as background real size, will be set again in its location.
#This procedure can de automatically invoked when saving an image by
#defining the AfterSaveImage event procedure like this:
#
#proc AfterSaveImage { filename type } {
#    GidUtils:SaveGeoreferenceFile $filename $type
#}
#
proc GidUtils:SaveGeoreferenceFile { filename type } {
    set fail 0
    switch $type {
	tif {
	    set newext tfw
	}
	jpg {
	    set newext jgw
	}
	png {
	    set newext pgw
	}
	gif {
	    set newext gfw
	}
	bmp {
	    set newext bpw
	}
	default {
	    set fail 1
	}
    }
    if { !$fail } {
	foreach item [GiD_Info view] {
	    #check that the current view is in XY plane, else don't save the georeference file
	    if { [lindex $item 0] == "m" } {
		for { set i 1 } { $i <= 12 } { incr i } {
		    set v [lindex $item $i]
		    switch $i {
		        1 -
		        6 -
		        11 {
		            if { [expr abs($v-1.0)] > 1e-6 } {
		                set fail 1
		                break
		            }
		        }
		        default {
		            if { [expr abs($v)] > 1e-6 } {
		                set fail 1
		                break
		            }
		        }
		    }
		}
		break
	    }
	}
	if { $fail } {
	    GidUtils::SetWarnLine [_ "Georeference file not saved because current view is not XY"]
	}
    }
    if { !$fail } {
	set newfilename [file rootname $filename].$newext
	set fp [open $newfilename w]
	if { $fp == "" } {
	    set fail 1
	} else {
	    foreach {L R B T N F E} [GiD_Info ortholimits] break
	    set Mxx [expr {double($R-$L)/[winfo width .gid.central.s]}]
	    set Myx 0
	    set Mxy 0
	    set Myy [expr {double($B-$T)/[winfo height .gid.central.s]}]
	    set Dx $L
	    set Dy $T
	    puts $fp $Mxx
	    puts $fp $Myx
	    puts $fp $Mxy
	    puts $fp $Myy
	    puts $fp $Dx
	    puts $fp $Dy
	    close $fp
	    GidUtils::SetWarnLine [_ "Georeference file saved"]
	}
    }
    return $fail
}


#to be used by TKWIDGET to easily add a select file button
#e.g.
#QUESTION: your_question
#VALUE: your_filename
#TKWIDGET: GidUtils::TkwidgetGetFilenameButton

proc GidUtils::TkwidgetGetFilenameButton { event args } {
    global tkwidgedprivfilenames tkwidgedprivfilebuttons
    switch $event {
	INIT {
	    set PARENT [lindex $args 0]
	    upvar [lindex $args 1] ROW
	    set GDN [lindex $args 2]
	    set STRUCT [lindex $args 3]
	    set QUESTION [lindex $args 4]
	    #initialize variable to current field value
	    set tkwidgedprivfilenames($QUESTION) [DWLocalGetValue $GDN $STRUCT $QUESTION]
	    #set entry $PARENT.e$ROW
	    set entry ""
	    foreach item [grid slaves $PARENT -row [expr $ROW-1]] {
		if { [winfo class $item] == "Entry"  || [winfo class $item] == "TEntry" } {
		    #assumed that it is the only entry of this row
		    set entry $item
		    break
		}
	    }
	    if { $entry != "" } {
		set tkwidgedprivfilebuttons($QUESTION) [button $PARENT.bfolder$QUESTION \
		        -image [GetImage "folder.gif"] \
		        -command [list GidUtils::_GetFilenameCmd tkwidgedprivfilenames($QUESTION) $entry 1]]
		grid $tkwidgedprivfilebuttons($QUESTION) -row [expr $ROW-1] -column 2 -sticky w
		grid configure $entry -sticky ew
	    }
	}
	SYNC {
	    #set GDN [lindex $args 0]
	    #set STRUCT [lindex $args 1]
	    #set QUESTION [lindex $args 2]
	    #DWLocalSetValue $GDN $STRUCT $QUESTION $tkwidgedprivfilenames($QUESTION)
	}
	DEPEND {
	    #set GDN [lindex $args 0]
	    #set STRUCT [lindex $args 1]
	    set QUESTION [lindex $args 2]
	    set ACTION [lindex $args 3]
	    #set value [lindex $args 4]
	    if { [info exists tkwidgedprivfilebuttons($QUESTION)] && \
		    [winfo exists $tkwidgedprivfilebuttons($QUESTION)] } {
		if { $ACTION == "HIDE" } {
		    grid remove $tkwidgedprivfilebuttons($QUESTION)
		} else {
		    #RESTORE
		    grid $tkwidgedprivfilebuttons($QUESTION)
		}
	    } else {

	    }
	}
	CLOSE {
	    array unset tkwidgedprivfilenames
	    array unset tkwidgedprivfilebuttons
	}
	default {
	    return [list ERROR [= "Unexpected tkwidget event"]]
	}
    }
    return ""
}


#axiliary procedure used by GidUtils::TkwidgetGetFilenameButton
proc GidUtils::_GetFilenameCmd { varname entry {tail 0}} {
    #set types [list [list [= " file"] ".dat"] [list [_ "All files"] ".*"]]
    #set defaultextension .dat
    set types [list [list [_ "All files"] ".*"]]
    set defaultextension ""
    set title [= "Select file"]
    set ::$varname [Browser-ramR file read .gid $title [set ::$varname] $types $defaultextension 0]
    if { [set ::$varname] != "" && $entry != "" && [winfo exists $entry] } {
	$entry delete 0 end
	if { $tail } {
	    $entry insert end [file tail [set ::$varname]]
	} else {
	    $entry insert end [set ::$varname]
	}
    }
    return [set ::$varname]
}

#to be used by TKWIDGET to easily add a pick node button
#e.g.
#QUESTION: your_question
#VALUE: your_node_id
#TKWIDGET: GidUtils::TkwidgetPickNode

proc GidUtils::TkwidgetPickNode { event args } {
    global tkwidgedprivpicknodebuttons
    switch $event {
	INIT {
	    set PARENT [lindex $args 0]
	    upvar [lindex $args 1] ROW
	    set GDN [lindex $args 2]
	    set STRUCT [lindex $args 3]
	    set QUESTION [lindex $args 4]
	    set entry ""
	    foreach item [grid slaves $PARENT -row [expr $ROW-1]] {
		if { [winfo class $item] == "Entry"  || [winfo class $item] == "TEntry" } {
		    #assumed that it is the only entry of this row
		    set entry $item
		    break
		}
	    }
	    set tkwidgedprivpicknodebuttons($QUESTION) [button $PARENT.bpicknode$QUESTION \
		    -image [GetImage "point16x16.gif"] -command [list GidUtils::_PickNodeCmd $entry]]
	    grid $tkwidgedprivpicknodebuttons($QUESTION) -row [expr $ROW-1] -column 2 -sticky w
	    grid configure $entry -sticky ew
	    return ""
	}
	SYNC {
	    #set GDN [lindex $args 0]
	    #set STRUCT [lindex $args 1]
	    #set QUESTION [lindex $args 2]
	    #DWLocalSetValue $GDN $STRUCT $QUESTION $value
	}
	DEPEND {
	    #set GDN [lindex $args 0]
	    #set STRUCT [lindex $args 1]
	    set QUESTION [lindex $args 2]
	    set ACTION [lindex $args 3]
	    #set value [lindex $args 4]
	    if { [info exists tkwidgedprivpicknodebuttons($QUESTION)] && \
		    [winfo exists $tkwidgedprivpicknodebuttons($QUESTION)] } {
		if { $ACTION == "HIDE" } {
		    grid remove $tkwidgedprivpicknodebuttons($QUESTION)
		} else {
		    #RESTORE
		    grid $tkwidgedprivpicknodebuttons($QUESTION)
		}
	    }
	}
	CLOSE {
	    array unset tkwidgedprivpicknodebuttons
	}
	default {
	    return [list ERROR [= "Unexpected tkwidget event"]]
	}
    }
}

#axiliary procedure used by GidUtils::TkwidgetPickNode
proc GidUtils::_PickNodeCmd { entry } {
    set numnode [GidUtils::PickEntities Nodes single]
    if { $numnode != "" } {
       $entry delete 0 end
       $entry insert end $numnode
    }
}

#to be used by TKWIDGET to easily add a layer combobox
#e.g.
#QUESTION: your_question
#VALUE: your_layername
#TKWIDGET: GidUtils::TkwidgetGetLayername

proc GidUtils::TkwidgetGetLayername { event args } {
    global tkwidgedprivlayer
    switch $event {
	INIT {
	    set PARENT [lindex $args 0]
	    upvar [lindex $args 1] ROW
	    set GDN [lindex $args 2]
	    set STRUCT [lindex $args 3]
	    set QUESTION [lindex $args 4]
	    #initialize variable to current field value
	    if { [DWLocalGetValue $GDN $STRUCT $QUESTION] != "" } {
		set tkwidgedprivlayer($QUESTION) [DWLocalGetValue $GDN $STRUCT $QUESTION]
	    } else {
		set tkwidgedprivlayer($QUESTION) [lindex [GiD_Info layers] 0]
	    }
	
	    #set entry $PARENT.e$ROW
	    set entry ""
	    foreach item [grid slaves $PARENT -row [expr $ROW-1]] {
		if { [winfo class $item] == "Entry"  || [winfo class $item] == "TEntry" } {
		    #assumed that it is the only entry of this row
		    set entry $item
		    break
		}
	    }
	    if { $entry != "" } {
		#set width [$entry cget -width]
		set width 9
		set tkwidgedprivfilebuttons($QUESTION) [ttk::combobox $PARENT.clayer \
		        -textvariable tkwidgedprivlayer($QUESTION) \
		        -values [GiD_Info Layers] \
		        -width $width]

		grid forget $entry
		grid $tkwidgedprivfilebuttons($QUESTION) -row [expr $ROW-1] -column 1 -sticky ew
	    }
	}
	SYNC {
	    set GDN [lindex $args 0]
	    set STRUCT [lindex $args 1]
	    set QUESTION [lindex $args 2]
	    DWLocalSetValue $GDN $STRUCT $QUESTION $tkwidgedprivlayer($QUESTION)
	}
	DEPEND {
	    #set GDN [lindex $args 0]
	    #set STRUCT [lindex $args 1]
	    set QUESTION [lindex $args 2]
	    set ACTION [lindex $args 3]
	    #set value [lindex $args 4]
	    if { [info exists tkwidgedprivlayer($QUESTION)] && \
		    [winfo exists $tkwidgedprivlayer($QUESTION)] } {
		if { $ACTION == "HIDE" } {
		    grid remove $tkwidgedprivlayer($QUESTION)
		} else {
		    #RESTORE
		    grid $tkwidgedprivlayer($QUESTION)
		}
	    } else {

	    }
	}
	CLOSE {
	    #array unset tkwidgedprivlayer
	}
	default {
	    return [list ERROR [= "Unexpected tkwidget event"]]
	}
    }
    return ""
}

proc GidUtils::IsTkDisabled { } {
    if { [info commands wm] == "" } {
	return 1
    } else {
	return 0
    }
}

proc GidUtils::AreWindowsDisabled { } {
    if { [GidUtils::IsTkDisabled] } {
	return 1
    }
    if { [winfo exists .gid.central.s] && [.gid.central.s disable windows] } {
	return 1
    }
    return 0
}


#window dialog examples
#
#get a filename
#to write
#   set w .gid ;#parent
#   set types [list [list [= "IGES file"] ".igs .iges"] [list [_ "All files"] ".*"]]
#   set defaultextension .igs
#   set title [= "Write IGES file"]
#   set filename [Browser-ramR file write $w $title "" $types $defaultextension 0]
#to read
#   set title [= "Read IGES file"]
#   set filename [Browser-ramR file read $w $title $filename $types $defaultextension 0]

